@@ -68,7 +68,7 @@ class Runner:
 			else:
 				return self.RETURNCODE_PROGRAM_NOT_FOUND
 
-	def run_and_get_output(self, cmd):
+	def run_and_get_output(self, cmd, *, ignore_returncode=False):
 		"""
 		Executes cmd without a shell and returns stdout as unicode/str.
 		Raises CommandError if program is not found or fails.
@@ -78,7 +78,7 @@ class Runner:
 		cmd = self._amend_command(cmd)
 
 		try:
-			p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
+			p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=not ignore_returncode)
 		except (CalledProcessError, FileNotFoundError) as e:
 			raise CommandError(cmd, e)
 		out = p.stdout
