@@ -1277,10 +1277,8 @@ class yank(commands.Command):
 
 		if hash_id is None:
 			raise ErrorInCommand(_("no commit selected"))
-		if hash_id in (self.app.ID_STAGED, self.app.ID_UNSTAGED, self.app.ID_UNTRACKED):
-			raise ErrorInCommand(_("uncommitted changes have no meta info which could be copied"))
-		if hash_id in (self.app.ID_STASHES_GROUP,):
-			raise ErrorInCommand(_("please open the group and select the stash you want to copy"))
+		if hash_id in self.app.SPECIAL_IDS:
+			raise ErrorInCommand(_("there is no meta info which could be copied"))
 
 		if type_id == self.app.TYPE_TAG:
 			sorted_placeholders = self.sorted_placeholders_tag
