@@ -14,23 +14,10 @@ from .string_formatter import formatter
 from .color_decoder import ColorDecoder
 from . import utils
 from . import settings
+from .constants import *
 
 ALIGN_CHAR = "^"
 
-ID_UNTRACKED = "<untracked>"
-ID_UNSTAGED = "<unstaged>"
-ID_STAGED = "<staged>"
-ID_STASHES_GROUP = "<stashed>"
-SPECIAL_IDS = (ID_UNTRACKED, ID_UNSTAGED, ID_STAGED, ID_STASHES_GROUP)
-
-TYPE_STASHED = "stashed"
-TYPE_OTHER = "other"
-TYPE_ERROR = "error"
-TYPE_TAG = "tag"
-TYPE_START_OF_FILE = "sof"
-TYPE_NUMBERED_LINE = "numbered-line"
-TYPE_UNTRACKED = "untracked"
-
 #https://matthew-brett.github.io/curious-git/git_object_types.html
 OBJECT_TYPE_COMMIT = "commit"
 OBJECT_TYPE_TREE = "tree"
@@ -45,6 +32,7 @@ title_unstaged = _('Local uncommitted changes, not checked in to index')
 title_unstaged_and_untracked = formatter.format(_('{title_unstaged} and {title_untracked:l}'), title_unstaged=title_unstaged, title_untracked=title_untracked)
 title_staged   = _('Local changes checked in to index but not committed')
 title_stashes_group = _('{number} stashes')
+title_todo = _('{number} TODO flags')
 title_referencedby = _('This commit is referenced by the following commits:')
 title_commit = _("commit {hash_id}{refnames}")
 pattern_refnames = _(" ({refnames})")
@@ -76,6 +64,17 @@ class LineInfo:
 		'formatted_linenumber_after', 'formatted_linenumber_before_right', 'formatted_linenumber_before_left',
 		'is_modified')
 
+	def __init__(self, lnno=None):
+		self.linenumber_after = lnno
+		self.linenumber_before_right = lnno
+		self.linenumber_before_left = lnno
+
+		self.formatted_linenumber_after = None
+		self.formatted_linenumber_before_right = None
+		self.formatted_linenumber_before_left = None
+
+		self.is_modified = None
+
 	def __repr__(self):
 		args = ", ".join("%s" % getattr(self, a) for a in ('linenumber_after', 'linenumber_before_right', 'linenumber_before_left'))
 		return "%s(%s)" % (type(self).__name__, args)
@@ -261,7 +260,7 @@ class LogModel(Model):
 	fmt_log    = "%Cred%h%Creset -^%C(yellow)%d%Creset %s %Cgreen(%cd) %C(bold blue)<%an>%Creset"
 	fmt_reflog = "%Cred%h %Creset%C(yellow)%gD%Creset^%C(yellow)%d%Creset %gs %Cgreen(%cd) %C(bold blue)<%gn>%Creset"
 
-	cmd_log = [["$status"], ["git", "log", "--color", "--graph", "--date=relative", FLAG_CUSTOM_FORMAT, "--abbrev-commit"]]
+	cmd_log = [["$stashes"], ["$todo"], ["$status"], ["git", "log", "--color", "--graph", "--date=relative", FLAG_CUSTOM_FORMAT, "--abbrev-commit"]]
 	cmd_stash_list = ["git", "stash", "list", "--color", "--pretty=format:%C(cyan)* %Cred%gd%Creset - ^%s %Cgreen(%cr)%Creset", "--abbrev-commit"]
 	cmd_unreachable = ["git", "fsck", "--unreachable"]
 
@@ -334,7 +333,7 @@ class LogModel(Model):
 	show_untracked = True
 	untracked_files_as_separate_group = False
 
-	show_stashed_changes = True
+	show_stashed_changes = True   # NOTE: deprecated, not needed anymore since commit fab472fb25359739418fb9ef1a5ba5d9d2480653
 	stashed_changes_reversed_order = False
 	stashed_changes_group = None  # True|False|None. None = auto
 
@@ -406,8 +405,30 @@ class LogModel(Model):
 			ln = self.format_log_entry(settings.COLOR_LOG_UNTRACKED, title_untracked)
 			log.insert(0, (ln, ID_UNTRACKED, TYPE_OTHER))
 
+		return log
+
+	status.arguments = ''
+	status.description = _('Show untracked, unstaged and staged changes if existing')
+	status.settings = [
+		'log-view.show-untracked',
+		'log-view.show-untracked-as-separate-group',
+		'log-view.line-pattern',
+		'title.untracked',
+		'title.unstaged',
+		'title.untracked-and-unstaged',
+		'title.staged',
+		settings.COLOR_LOG_UNTRACKED,
+		settings.COLOR_LOG_UNSTAGED,
+		settings.COLOR_LOG_STAGED,
+	]
+
+	def stashes(self, args):
+		if args:
+			cmd = ['$stashes'] + args
+			raise CommandError(cmd, "stashes takes no arguments")
+
+		log = []
 
-		# stashed changes
 		if self.show_stashed_changes:
 			self.cmd = self.cmd_stash_list
 			try:
@@ -426,29 +447,53 @@ class LogModel(Model):
 
 		return log
 
-	status.arguments = ''
-	status.description = _('Show stashed, untracked, unstaged and staged changes if existing')
-	status.settings = [
-		'log-view.show-untracked',
-		'log-view.show-untracked-as-separate-group',
+	stashes.arguments = ''
+	stashes.description = _('Show stashed changes if existing')
+	stashes.settings = [
 		'log-view.cmd.stash-list',
 		'log-view.show-stashes',
 		'log-view.show-stashes-as-group',
 		'log-view.show-stashes-in-reverse-order',
 		'log-view.line-pattern',
 		'title.stashes',
-		'title.untracked',
-		'title.unstaged',
-		'title.untracked-and-unstaged',
-		'title.staged',
 		settings.COLOR_LOG_STASHES,
-		settings.COLOR_LOG_UNTRACKED,
-		settings.COLOR_LOG_UNSTAGED,
-		settings.COLOR_LOG_STAGED,
+	]
+
+	def todo(self, args):
+		if args:
+			cmd = ['$todo'] + args
+			raise CommandError(cmd, "todo takes no arguments")
+
+		root = get_git_root()
+		self.cmd = self.replace_in_cmd(DetailsModel.cmd_grep_todo, settings.WC_GIT_ROOT, root)
+		try:
+			lines = super().get_lines()
+		except CommandError as e:
+			if e.returncode == 1:
+				# no todo flags found
+				return []
+			return self.format_command_error(e, setting='details-view.cmd.todo.grep')
+		else:
+			if not lines:
+				return []
+
+			ln = formatter.format(title_todo, number=len(lines))
+			ln = self.format_log_entry(settings.COLOR_LOG_TODO, ln)
+			return [(ln, ID_TODO, TYPE_TODO)]
+
+	todo.arguments = ''
+	todo.description = _('Show TODO flags')
+	todo.settings = [
+		'details-view.cmd.todo.grep',
+		'log-view.line-pattern',
+		'title.todo',
+		settings.COLOR_LOG_TODO,
 	]
 
 	command_functions = dict(
 		status = status,
+		stashes = stashes,
+		todo = todo,
 	)
 
 	def format_log_entry(self, color, logentry):
@@ -546,6 +591,8 @@ class DetailsModel(Model):
 	cmd_unstaged = ["git", "diff", "--color"]
 	cmd_staged = cmd_unstaged + ["--cached"]
 	cmd_untracked = ["$untracked"]
+	cmd_todo = ["$todo"]
+	cmd_grep_todo = ["git", "grep", "--no-index", "--exclude-standard", "-n", "-I", "-E", '(#|%|"|//)TODO\>|\<TODO[:]', '--', settings.WC_GIT_ROOT]
 	cmd_pattern_stashed = [
 			["git", "log", "-g", "-1", "--pretty=format:%C(yellow)%gD%Creset%ncommit %H%nAuthor: %an <%ae>%nDate:   %ad%w(,4,4)%n%+B%n%n", "--color", WC_HASH_ID],
 			["git", "stash", "show", "--patch", "--color", WC_HASH_ID],
@@ -648,6 +695,7 @@ class DetailsModel(Model):
 	ln_pattern_untracked_in_separate_group = _("{filename}")
 
 	pattern_user = _("{name} <{email}>")
+	pattern_todo = _("<color=magenta>{filename}</color><color=cyan>:</color><color=green>{linenumber}</color><color=cyan>:</color>{todoflag}")
 
 	linenumber_new = "{new}"
 	linenumber_old = "{old}"
@@ -679,6 +727,7 @@ class DetailsModel(Model):
 	reo_url = re.compile(r'\S+?([.]\S+)+')
 
 	untracked_relative = settings.RELATIVE_CWD
+	todo_relative = settings.RELATIVE_NAME_ONLY
 
 	# if linenumbers are activated: insert ALIGN_CHAR in every line which is not represented by a tuple
 	insert_align_character = True
@@ -707,6 +756,10 @@ class DetailsModel(Model):
 			self.cmd = self.cmd_untracked
 			self.setting_cmd = 'details-view.cmd.untracked'
 			self.title = [(settings.COLOR_TITLE, title_untracked)]
+		elif self.hash_id == ID_TODO:
+			self.cmd = self.cmd_todo
+			self.setting_cmd = 'details-view.cmd.todo'
+			self.title = None
 		elif self.id_type == TYPE_STASHED:
 			self.cmd = self.replace_in_cmd(self.cmd_pattern_stashed, self.WC_HASH_ID, self.hash_id)
 			self.setting_cmd = 'details-view.cmd.stash'
@@ -744,7 +797,8 @@ class DetailsModel(Model):
 		- [("attr", "text")]  # urwid markup, must not contain tabs!
 		- ("text", type)
 		- ("text", type, auto_move_align_character)
-		- ("text", TYPE_UNTRACKED, file_name)
+		- ("text", TYPE_UNTRACKED, file_info)
+		- ("text", TYPE_TODO, file_info, line_info)
 		- ("text", TYPE_START_OF_FILE, file_info)
 		- ("text", TYPE_NUMBERED_LINE, auto_move_align_character, line_info)
 		"""
@@ -1011,6 +1065,71 @@ class DetailsModel(Model):
 
 	# ------- command functions ------
 
+	todo_strip_indentation = True
+	todo_only = False
+
+	def todo(self, args):
+		if args:
+			cmd = ['$todo'] + args
+			raise CommandError(cmd, "todo takes no arguments")
+
+		root = get_git_root()
+		self.cmd = self.replace_in_cmd(self.cmd_grep_todo, settings.WC_GIT_ROOT, root)
+		try:
+			lines = super().get_lines()
+		except CommandError as e:
+			if e.returncode == 1:
+				# no todo flags found
+				return [[(settings.COLOR_TITLE, title_todo.format(number=0))]]
+			return self.format_command_error(e, setting='details-view.cmd.todo.grep')
+		else:
+			out = []
+			n = 0
+			for ln in lines:
+				ln = ln.replace('\t', self.app.DetailsView.TAB)
+				try:
+					filename, linenumber, todoflag = ln.split(':', 2)
+					linenumber = int(linenumber)
+				except ValueError:
+					out.append(ln)
+				else:
+					filename = self.decode_filename(filename)
+					if filename.startswith('"') and filename.endswith('"'):
+						filename = filename[1:-1]
+					absfilename = os.path.abspath(filename)
+
+					if self.todo_relative == settings.RELATIVE_NAME_ONLY:
+						filename = os.path.split(filename)[1]
+					elif self.todo_relative == settings.RELATIVE_ROOT:
+						filename = absfilename
+					elif self.todo_relative == settings.RELATIVE_CWD:
+						pass
+					elif self.todo_relative == settings.RELATIVE_GIT:
+						filename = os.path.relpath(absfilename, root)
+					else:
+						assert False, 'invalid value for self.todo_relative: %r' % self.todo_relative
+
+					if self.todo_strip_indentation:
+						todoflag = todoflag.strip()
+					if self.todo_only:
+						i = todoflag.index('TODO') + 4
+						todoflag = todoflag[i:].lstrip(':').lstrip()
+
+					markup = formatter.format(self.pattern_todo, filename=filename, linenumber=linenumber, todoflag=todoflag)
+					markup = utils.colored_str_to_markup(markup, self.app.define_color)
+					out.append((markup, TYPE_TODO, FileInfo(absfilename), LineInfo(linenumber)))
+					n += 1
+			out.insert(0, [(settings.COLOR_TITLE, title_todo.format(number=n))])
+			out.insert(1, '')
+			return out
+
+	todo.arguments = ''
+	todo.description = _('Show a list of the TODO flags in the code')
+	todo.settings = [
+		'details-view.cmd.todo.grep',
+		'details-view.todo.line-pattern'
+	]
+
 	def referencedby(self, args):
 		assert len(args) == 1
 		hash_id = args[0]
@@ -1205,6 +1324,7 @@ class DetailsModel(Model):
 		referencedby = referencedby,
 		untracked = untracked,
 		commitheader = commitheader,
+		todo = todo,
 	)
 
 
