@@ -5,11 +5,13 @@ import gettext
 _ = gettext.gettext
 
 from .urwid_text_layout import LogTextLayout as TextLayout
+from .constants import SPECIAL_DETAILS_IDS, VIRTUAL_ID_OTHER
 
 ATTR_SEP = '.'
 GROUP_SEP = '.'
 
 WC_HASH_ID = "{hash}"
+WC_GIT_ROOT = "{git-root}"
 
 BOOL_TRUE = "true"
 BOOL_FALSE = "false"
@@ -63,6 +65,7 @@ COLOR_LOG_STAGED = "color.log-entry.staged"
 COLOR_LOG_UNSTAGED = "color.log-entry.unstaged"
 COLOR_LOG_UNTRACKED = "color.log-entry.untracked"
 COLOR_LOG_STASHES = "color.log-entry.stashes"
+COLOR_LOG_TODO = "color.log-entry.todo"
 COLOR_DETAILS_UNTRACKED = "color.details-view.line-pattern-untracked"
 COLOR_LINE_NUMBERS = "color.linenumber"
 COLORS = (
@@ -74,7 +77,7 @@ COLORS = (
 	COLOR_ELLIPSIS, COLOR_SPECIAL_CHARACTER,
 	COLOR_KEY_PRESSED, COLOR_KEY, COLOR_CMD, COLOR_MNEMONIC,
 	COLOR_CMD_IN_TEXT, COLOR_KEY_IN_TEXT, COLOR_CMD_IN_LINK,
-	COLOR_LOG_STAGED, COLOR_LOG_UNSTAGED, COLOR_LOG_UNTRACKED, COLOR_LOG_STASHES,
+	COLOR_LOG_STAGED, COLOR_LOG_UNSTAGED, COLOR_LOG_UNTRACKED, COLOR_LOG_STASHES, COLOR_LOG_TODO,
 	COLOR_DETAILS_UNTRACKED,
 	COLOR_LINE_NUMBERS,
 )
@@ -105,6 +108,7 @@ LOG_LEVELS = (
 RELATIVE_ROOT = 'root'
 RELATIVE_GIT = 'git'
 RELATIVE_CWD = 'cwd'
+RELATIVE_NAME_ONLY = 'file-name-only'
 
 relative_values = (
 	(RELATIVE_CWD, RELATIVE_CWD, _("Paths relative to current working directory")),
@@ -112,6 +116,10 @@ relative_values = (
 	(RELATIVE_ROOT, RELATIVE_ROOT, _("Absolute paths")),
 )
 
+relative_values_todo = relative_values + (
+	(RELATIVE_NAME_ONLY, RELATIVE_NAME_ONLY, _("File names only, without any directories")),
+)
+
 bool_values = ((BOOL_TRUE, True), (BOOL_FALSE, False))
 align_values = (
 	("left-indentation", TextLayout.ALIGN_INDENTATION, _("Left align but add an indentation to broken lines for better readability. The amount of indentation is specified by the align character %app.align-character%. The align character is not displayed and is automatically removed or restored from the commands when changing this option. If this character occurs several times in a line only the first occurence is treated as align character and the others are displayed normally. If a line contains no align character it is assumed to be present before the first character. I have implemented this align mode myself. If it causes trouble try urwid's default left.")),
@@ -140,7 +148,7 @@ settings = {
 		(BOOL_TRUE,  True,  _("Show untracked files separate from unstaged changes.")),
 		(BOOL_FALSE, False, _("Show untracked files together with unstaged changes.")),
 	), _("This setting is ignored if %log-view.show-untracked% is disabled.")),
-	"log-view.show-stashes" : ("app.LogModel.show_stashed_changes", bool, _("Show stashed changes (if existing) in the log.")),
+	"log-view.show-stashes" : ("app.LogModel.show_stashed_changes", bool, _("Show stashed changes (if existing) in the log. This setting is deprecated. Add or remove $stashes from %log-view.cmd.log% instead.")),
 	"log-view.show-stashes-as-group" : ("app.LogModel.stashed_changes_group", (
 		(BOOL_TRUE,  True,  _("Group all stashes to one list item. Opening that list item opens a list of all stashes.")),
 		(BOOL_FALSE, False, _("Show one list item for each stash.")),
@@ -163,7 +171,8 @@ settings = {
 	"details-view.indent-broken-code" : ("app.DetailsModel.indent_broken_code", bool, "Move the align character to the next non-whitespace character (this requires align=left-indentation and can be influenced by %app.tab%)"),
 	"details-view.max-lines-per-file" : ("app.DetailsView.max_lines_per_file", int, "Trim long diffs to reduce loading time."),
 	"details-view.max-lines-per-file.blob" : ("app.DetailsView.max_lines_per_file_blob", int, "Trim long blobs to reduce loading time."),
-	"details-view.show-untracked-relative-to" : ("app.DetailsModel.untracked_relative", relative_values, _("How to show untracked files")),
+	"details-view.show-untracked-relative-to" : ("app.DetailsModel.untracked_relative", relative_values, _("How to show untracked files, see also %details-view.show-todo-relative-to%")),
+	"details-view.show-todo-relative-to" : ("app.DetailsModel.todo_relative", relative_values_todo, _("How to show file names returned by the $todo function, see also %details-view.show-untracked-relative-to%")),
 	"details-view.cmd.commit" : ("app.DetailsModel.cmd_pattern", TYPE_COMMAND, _("The command which generates the details view for a commit, should include the wild card {wc_hash}").format(wc_hash=WC_HASH_ID)),
 	"details-view.cmd.tag" : ("app.DetailsModel.cmd_pattern_tag", TYPE_COMMAND, _("The command which generates the details view for a tag, should include the wild card {wc_hash}").format(wc_hash=WC_HASH_ID)),
 	"details-view.cmd.blob" : ("app.DetailsModel.cmd_pattern_blob", TYPE_COMMAND, _("The command which generates the details view for a blob, should include the wild card {wc_hash}").format(wc_hash=WC_HASH_ID)),
@@ -172,6 +181,8 @@ settings = {
 	"details-view.cmd.staged" : ("app.DetailsModel.cmd_staged", TYPE_COMMAND, _("The command which generates the details view for staged changes")),
 	"details-view.cmd.unstaged" : ("app.DetailsModel.cmd_unstaged", TYPE_COMMAND, _("The command which generates the details view for unstaged changes")),
 	"details-view.cmd.untracked" : ("app.DetailsModel.cmd_untracked", TYPE_COMMAND, _("The command which generates the details view for untracked changes, each line is formatted with %details-view.unstaged.line-pattern-untracked% or %details-view.untracked.line-pattern-untracked% and colored with %color.details-view.line-pattern-untracked%")),
+	"details-view.cmd.todo" : ("app.DetailsModel.cmd_todo", TYPE_COMMAND, _("The command which generates the details view for TODO flags")),
+	"details-view.cmd.todo.grep" : ("app.DetailsModel.cmd_grep_todo", TYPE_COMMAND, _("The command used by the function $todo to generate the list of TODO flags, each line represents one TODO flag and has the pattern <filename>:<linenumber>:<linecontent>. The lines will be formatted with %details-view.todo.line-pattern%. May contain the wildcard {wc}.").format(wc=WC_GIT_ROOT)),
 	"details-view.cmd.diff" : ("app.DiffModel.cmd", TYPE_COMMAND, _("The beginning of the command which generates the content for gitd")),
 	"details-view.cmd.commit.header-data" : ("app.DetailsModel.cmd_start_header_data", TYPE_COMMAND, _("The beginning of the command which gathers the data for the function $commitheader, format and commit hash are appended automatically")),
 	"details-view.cmd.commit.referenced-by" : ("app.DetailsModel.cmd_referencing_commits", TYPE_COMMAND, _("The beginning of the command which generates the list of commits referencing the current commit in the function $referencedby, --grep is added automatically, --custom-format is replaced by %log-view.fmt.log%, the output is formatted by %details-view.commit.line-pattern-referenced-by%")),
@@ -180,9 +191,13 @@ settings = {
 	"details-view.commit.line-pattern-referenced-by" : ("app.DetailsModel.ln_pattern_referencedby", str, _("How to format an entry in the list of commits which reference the current commit, should include the wildcard {logentry}, this is related to %log-view.line-pattern% but does not support color markup, this should not contain an %app.align-character% if %details-view.cmd.commit.referenced-by% contains one")),
 	"details-view.commit.user-pattern" : ("app.DetailsModel.pattern_user", TYPE_COLORED_STR, _("How to display a user in $commitheader, supported are the wild cards {name} and {email}")),
 	"details-view.commit.indentation-commit-message" : ("app.DetailsModel.indent_body", str, _("An indentation for the commit message in $commitheader")),
+	"details-view.todo.line-pattern" : ("app.DetailsModel.pattern_todo", TYPE_COLORED_STR, _("How the $todo function formats TODO flags, may include the wildcards {filename}, {linenumber} and {todoflag}")),
+	"details-view.todo.strip-indentation" : ("app.DetailsModel.todo_strip_indentation", bool, _("Strip the indentation from the lines displayed by the $todo function")),
+	"details-view.todo.todo-only" : ("app.DetailsModel.todo_only", bool, _("Strip everything before the TODO from the lines displayed by the $todo function")),
 	"details-view.line-numbers.min-width" : ("app.DetailsModel.min_line_number_width", int, _("The amount of space to reserve for line numbers is determined for each hunk individually and depends on the biggest line number, this setting can be used to always reserve space for at least a certain number of digits so that the line numbers column has the same width for all hunks")),
 	"details-view.line-numbers.sep" : ("app.DetailsModel.linenumber_suffix", TYPE_COLORED_STR, _("A separator between the line number and the line itself")),
 	"details-view.line-numbers.sep-left-right" : ("app.DetailsModel.linenumber_sep", TYPE_COLORED_STR, _("A separator between the left and right line number (when showing the line numbers before a merge commit)")),
+	"details-view.auto-enable-visual" : ("app.DetailsView.visual_ids", (list, SPECIAL_DETAILS_IDS + (VIRTUAL_ID_OTHER,)), _("Automatically enable visual mode if a details view for this type of information is opened")),
 
 	#TODO: after changing these settings n and ? are not updated automatically, a new search must be started
 	"search.case-sensitive" : ("app.search_case_sensitive", bool_values + ((VAL_AUTO, None, _("Case sensitive if search text contains upper case characters, case insensitive otherwise")),), ""),
@@ -219,6 +234,7 @@ settings = {
 	"title.untracked-and-unstaged" : ("app.model.title_unstaged_and_untracked", str, _("Log entry and details view title for a combination of untracked files and unstaged changes (if %log-view.show-untracked-as-separate-group% is false and %log-view.show-untracked% is true)")),
 	"title.staged" : ("app.model.title_staged", str, _("Log entry and details view title for staged changes")),
 	"title.stashes" : ("app.model.title_stashes_group", str, _("Log entry and details view title for the list of all stashes (if %log-view.show-stashes% and %log-view.show-stashes-as-group% are not false)")),
+	"title.todo" : ("app.model.title_todo", str, _("Log entry and details view title for the list of TODO flags, may contain the wildcard {number}")),
 	# this is used in details-view only but seperating it from the other titles would not make sense
 	"title.commit" : ("app.model.title_commit", str, _("Title displayed in the details view for commits, the following wild cards are supported: {hash_id}, {refnames} (see %title.commit.refnames%), {subject}, {committer_name}, {committer_email}, {committer_date}, {author_name}, {author_email}, {author_date}, {body}")),
 	"title.commit.refnames" : ("app.model.pattern_refnames", str, _("A pattern how to format {refnames} in %title.commit% if the commit has ref names (decorations), this pattern should include the wild card {refnames}")),
